#include <config.h>

#include "th10.h"
#include "thcrypt.h"
#include "thlzss.h"

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <errno.h>

typedef struct {
	char* name;
	uint32_t name_length;
	uint32_t size;  /* uncompressed size of file */
	uint32_t zsize; /* compressed size of file */
} th10dat_entry;

typedef struct {
	FILE* file;
	unsigned int data_size;     /* Total size of the uncompressed files. */
	unsigned int data_zsize;    /* Total size of the compressed files. */
	th10dat_entry* entries;     /* The actual entries. */
	unsigned int entries_count; /* Number of entries. */
	unsigned int entries_size;  /* Total size of the file directory. */
	unsigned int entries_zsize; /* Total size of the compressed file directory. */
} th10dat_private;

static void
th10dat_private_free(th10dat_private* private)
{
	if (private) {
		if (private->entries) {
			unsigned int i;
			for (i = 0; i < private->entries_count; ++i) {
				if (private->entries[i].name)
					free(private->entries[i].name);
			}
			free(private->entries);
		}
		free(private);
	}
}

static void*
th10_open(FILE* fd)
{
	th10dat_private* private;

	if (!fd) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "fd is NULL");
		return NULL;
	}

	/* Write the header later. */
	if (fseek(fd, 16, SEEK_SET) == -1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't seek: %s", strerror(errno));
		return NULL;
	}

	private = malloc(sizeof(th10dat_private));
	if (!private) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "malloc failed: %s", strerror(errno));
		return NULL;
	}

	private->file = fd;

	private->data_size = 0;
	private->data_zsize = 0;
	private->entries = NULL;
	private->entries_count = 0;
	private->entries_size = 0;
	private->entries_zsize = 0;

	return private;
}

static int
th10_write(th10dat_private* private, FILE* fd, const char* filename)
{
	unsigned int i;
	th10dat_entry* entry;
	char* path;
	char* name;
	unsigned int namelen;
	unsigned char* xdata;
	unsigned char* zdata;
	if (!private) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "private is NULL");
		return -1;
	}

	if (!fd || !filename) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "fd or filename is NULL");
		th10dat_private_free(private);
		return -1;
	}

	path = strdup(filename);
	name = basename(path);

	if (private->entries) {
		for (i = 0; i < private->entries_count; ++i) {
			if (strcmp(name, private->entries[i].name) == 0) {
				snprintf(library_error, LIBRARY_ERROR_SIZE, "duplicate filename %s", name);
				free(path);
				th10dat_private_free(private);
				return -1;
			}
		}
	}

	private->entries_count++;
	private->entries = realloc(private->entries, sizeof(th10dat_entry) * private->entries_count);
	entry = &private->entries[private->entries_count - 1];

	namelen = strlen(name) + 4;
	namelen -= namelen % 4;
	entry->name = malloc(namelen);
	if (!entry->name) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "malloc failed: %s", strerror(errno));
		free(path);
		th10dat_private_free(private);
		return -1;
	}
	memset(entry->name, 0, namelen);
	memcpy(entry->name, name, strlen(name));

	private->entries_size += namelen + 12;
	entry->name_length = namelen;
	entry->size = 0;
	entry->zsize = 0;

	/* Find out file size, might be better to use stat, but it's probably less cross-platform. */
	/* This will fail for 2GB+ files, but who cares. */
	if (fseek(fd, 0, SEEK_END) == -1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't seek: %s", strerror(errno));
		free(path);
		th10dat_private_free(private);
		return -1;
	}

	entry->size = ftell(fd);
	if (entry->size == -1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "ftell failed: %s", strerror(errno));
		free(path);
		th10dat_private_free(private);
		return -1;
	}
	private->data_size += entry->size;

	if (fseek(fd, 0, SEEK_SET) == -1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't seek: %s", strerror(errno));
		free(path);
		th10dat_private_free(private);
		return -1;
	}

	zdata = th_lz_fd(fd, &entry->zsize);
	if (!zdata) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "th_lz_fd failed (TODO: Improve error message.)");
		free(path);
		th10dat_private_free(private);
		return -1;
	}
	if (entry->zsize >= entry->size) {
		if (fseek(fd, 0, SEEK_SET) == -1) {
			snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't seek: %s", strerror(errno));
			free(zdata);
			free(path);
			th10dat_private_free(private);
			return -1;
		}

		if (fread(zdata, entry->size, 1, fd) != 1) {
			snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't read: %s", strerror(errno));
			free(zdata);
			free(path);
			th10dat_private_free(private);
			return -1;
		}

		entry->zsize = entry->size;
	}

	private->data_zsize += entry->zsize;

	xdata = malloc(entry->zsize);
	if (!xdata) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "malloc failed: %s", strerror(errno));
		free(zdata);
		free(path);
		th10dat_private_free(private);
		return -1;
	}

	i = th_get_cryprm_index(name);
	free(path);
	th_encrypt(zdata, xdata, entry->zsize, g_cryprm[i].key, g_cryprm[i].step, g_cryprm[i].block, g_cryprm[i].limit);
	free(zdata);

	if (fwrite(xdata, entry->zsize, 1, private->file) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't write: %s", strerror(errno));
		free(xdata);
		free(path);
		th10dat_private_free(private);
		return -1;
	}
	free(xdata);

	return 0;
}

static int
th10_close(th10dat_private* private)
{
	unsigned char* buffer;
	unsigned char* buffer_ptr;
	unsigned int i;
	/* XXX: Why 16? */
	uint32_t offset = 16;
	unsigned char* zbuffer;
	unsigned char* xbuffer;
	uint32_t header[4];
	unsigned char xheader[sizeof(header)];

	if (!private) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "private is NULL");
		return -1;
	}

	/* Write entry directory. */
	buffer = malloc(private->entries_size);
	if (!buffer) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "malloc failed: %s", strerror(errno));
		th10dat_private_free(private);
		return -1;
	}
	memset(buffer, 0, private->entries_size);

	buffer_ptr = buffer;
	for (i = 0; i < private->entries_count; ++i) {
		th10dat_entry* entry = &private->entries[i];
		memcpy(buffer_ptr, entry->name, entry->name_length);
		memcpy(buffer_ptr + entry->name_length, &offset, sizeof(uint32_t));
		memcpy(buffer_ptr + entry->name_length + 4, &entry->size, sizeof(uint32_t));
		buffer_ptr += entry->name_length + 12;
		offset += entry->zsize;
	}

	zbuffer = th_lz_mem(buffer, private->entries_size, &private->entries_zsize);
	if (!zbuffer) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "th_lz_mem failed (TODO: Improve error message.)");
		free(buffer);
		th10dat_private_free(private);
		return -1;
	}
	free(buffer);

	xbuffer = malloc(private->entries_zsize);
	if (!xbuffer) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "malloc failed: %s", strerror(errno));
		free(zbuffer);
		th10dat_private_free(private);
		return -1;
	}
	th_encrypt(zbuffer, xbuffer, private->entries_zsize, 0x3e, 0x9b, 0x80, private->entries_size);
	free(zbuffer);

	if (fwrite(xbuffer, private->entries_zsize, 1, private->file) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't write: %s", strerror(errno));
		free(xbuffer);
		th10dat_private_free(private);
		return -1;
	}
	free(xbuffer);

	/* Write header. */
	header[0] = 0x31414854;
	header[1] = private->entries_size + 123456789;
	header[2] = private->entries_zsize + 987654321;
	header[3] = private->entries_count - 0xf7e7f8ac;

	th_encrypt((unsigned char*)&header, xheader, sizeof(header), 0x1b, 0x37, sizeof(header), sizeof(header));

	if (fseek(private->file, 0, SEEK_SET) == -1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't seek: %s", strerror(errno));
		th10dat_private_free(private);
		return -1;
	}
	if (fwrite(&xheader, sizeof(header), 1, private->file) != 1) {
		snprintf(library_error, LIBRARY_ERROR_SIZE, "couldn't write: %s", strerror(errno));
		th10dat_private_free(private);
		return -1;
	}

	th10dat_private_free(private);

	return 0;
}

thdat_fptrs th10dat_fptrs = {
	th10_open,
	(int (*)(void*, FILE*, const char*))th10_write,
	(int (*)(void*))th10_close
};
