#include <config.h>
#include <stdio.h>
/* realloc, abort */
#include <stdlib.h>

/* Compression specification:
 *
 * The LZSS algorithm is used.  A single byte is used in front of every entry:
 * 1 means that a byte follows, and 0 that an (offset, length) follows. The
 * size of the offset is 13 bits, and the length is 4 bits.  The minimum match
 * length is 3, allowing for a maximum match length of 18.  Offsets cannot
 * refer to entry 0 in the dictionary.
 *
 * The dictionary is 8192 bytes large, writing must start at index 1.  The last
 * 18 bytes (the dictionary pointer and in front of it) in the dictionary
 * cannot be read.  The dictionary is initialized to 0.
 *
 * Forward-looking matches are possible for repeating data. */

typedef struct {
	unsigned char* buffer;
	unsigned int buffer_size;
	unsigned int used_bytes;

	unsigned char next_byte;
	unsigned char used_bits;
} bitstream_t;

static void
bitstream_grow(bitstream_t* b)
{
	b->buffer_size <<= 1;
	b->buffer = realloc(b->buffer, b->buffer_size);
	if (!b->buffer)
		abort();
}

static void
bitstream_init(bitstream_t* b)
{
	b->used_bytes = 0;
	b->buffer = NULL;
	b->buffer_size = 2048; /* Grows to 4096 right away. */
	b->next_byte = 0;
	b->used_bits = 0;
	bitstream_grow(b);
}

static void
bitstream_write1(bitstream_t* b, unsigned int bit)
{
	b->next_byte |= bit << (7 - b->used_bits);
	b->used_bits++;
	if (b->used_bits >= 8) {
		b->buffer[b->used_bytes] = b->next_byte;
		b->used_bytes++;
		if (b->used_bytes >= b->buffer_size)
			bitstream_grow(b);
		b->next_byte = 0;
		b->used_bits = 0;
	}
}

static void
bitstream_finish(bitstream_t* b)
{
	while (b->used_bits)
		bitstream_write1(b, 0);
}

/* Can write at most ``unsigned int'' bits. */
static void
bitstream_write(bitstream_t* b, unsigned int bits, unsigned int data)
{
	int i;
	for (i = bits - 1; i >= 0; --i) {
		const unsigned int bit = (data >> i) & 1;
		bitstream_write1(b, bit);
	}
}

#define LZSS_DICTSIZE 0x2000
#define LZSS_MIN_MATCH 3
#define LZSS_MAX_MATCH (LZSS_MIN_MATCH + 15)

#define HASH_SIZE 8192
#define HASH_d    5
#define HASH_NULL 0

typedef struct {
	/* These are all indices into prev or next. */
	unsigned int hash[HASH_SIZE];
	unsigned int prev[LZSS_DICTSIZE];
	unsigned int next[LZSS_DICTSIZE];
} hash_t;

static unsigned int
generate_key(const unsigned char* array)
{
	unsigned int key = 0;
	unsigned int i;
	for (i = 0; i < LZSS_MIN_MATCH; ++i) {
		key = (key << HASH_d) ^ (array[i]);
		key %= HASH_SIZE;
	}
	return key;
}

static unsigned int
generate_key_safe(const unsigned char* array, unsigned int base, unsigned int wrap)
{
	unsigned int key = 0;
	unsigned int i;
	for (i = 0; i < LZSS_MIN_MATCH; ++i) {
		key = (key << HASH_d) ^ (array[(base + i) % wrap]);
		key %= HASH_SIZE;
	}
	return key;
}

static void
list_remove(hash_t* hash, const unsigned int key, const unsigned int offset)
{
	const unsigned int P = hash->prev[offset];
	const unsigned int N = hash->next[offset];

	if (P == HASH_NULL)
		hash->hash[key] = N;

	if (N != HASH_NULL)
		hash->prev[N] = P;

	if (P != HASH_NULL)
		hash->next[P] = N;

	hash->prev[offset] = HASH_NULL;
	hash->next[offset] = HASH_NULL;
}

static void
list_add(hash_t* hash, const unsigned int key, const unsigned int offset)
{
	if (hash->hash[key] == HASH_NULL) {
		hash->hash[key] = offset;
		hash->next[offset] = HASH_NULL;
		hash->prev[offset] = HASH_NULL;
	} else {
		hash->next[offset] = hash->hash[key];
		hash->prev[hash->hash[key]] = offset;
		hash->prev[offset] = HASH_NULL;
		hash->hash[key] = offset;
	}
}

/*unsigned char*
th_lz0(const unsigned char* in, unsigned int insize, unsigned int* outsize)
{
	bitstream_t bs;
	const unsigned char* in_ptr = in;

	bitstream_init(&bs);

	while (in_ptr - in < insize) {
		bitstream_write1(&bs, 1);
		bitstream_write(&bs, 8, *in_ptr++);
	}

	bitstream_finish(&bs);
	*outsize = bs.used_bytes;
	return bs.buffer;
}*/

typedef int (*read_byte_fptr)(void*);

static unsigned char*
th_lz(unsigned int* outsize, read_byte_fptr read_byte, void* data)
{
	bitstream_t bs;
	hash_t hash;
	unsigned char dict[LZSS_DICTSIZE];
	/* XXX: Magic. */
	unsigned int dict_head = 1;
	unsigned int waiting_bytes = 0;
	unsigned int c, i;

	bitstream_init(&bs);
	for (i = 0; i < HASH_SIZE; ++i)
		hash.hash[i] = HASH_NULL;

	for (i = 0; i < LZSS_DICTSIZE; ++i) {
		hash.next[i] = hash.prev[i] = HASH_NULL;
		dict[i] = 0;
	}

	/* Fill the forward-looking buffer. */
	for (i = 0; i < LZSS_MAX_MATCH; ++i) {
		c = read_byte(data);
		if (c == -1)
			break;
		dict[dict_head + i] = c;
		waiting_bytes++;
	}

	while (waiting_bytes) {
		unsigned int match_len = LZSS_MIN_MATCH - 1;
		unsigned int match_offset = 0;
		unsigned int offset;

		/* Find the best match. */
		for (offset = hash.hash[generate_key_safe(dict, dict_head, LZSS_DICTSIZE)]; offset != HASH_NULL; offset = hash.next[offset]) {
			unsigned int match_tmp = 0;
			for (i = 0; i < waiting_bytes; ++i) {
				if (dict[(dict_head + i) % LZSS_DICTSIZE] != dict[(offset + i) % LZSS_DICTSIZE])
					break;
				match_tmp++;
			}
			if (match_tmp > match_len) {
				match_len = match_tmp;
				match_offset = offset;
				if (match_len == waiting_bytes)
					break;
			}
		}

		/* Write data to the output buffer. */
		if (match_len < LZSS_MIN_MATCH) {
			match_len = 1;
			bitstream_write1(&bs, 1);
			bitstream_write(&bs, 8, dict[dict_head]);
		} else {
			bitstream_write1(&bs, 0);
			bitstream_write(&bs, 13, match_offset);
			bitstream_write(&bs, 4, match_len - LZSS_MIN_MATCH);
		}

		/* Add bytes to the dictionary. */
		for (i = 0; i < match_len; ++i) {
			unsigned int offset = (dict_head + LZSS_MAX_MATCH) % LZSS_DICTSIZE;
			if (offset != 0)
				list_remove(&hash, generate_key_safe(dict, offset, LZSS_DICTSIZE), offset);

			c = read_byte(data);
			if (c != -1) {
				dict[offset] = c;
				waiting_bytes++;
			}

			if (dict_head != 0)
				list_add(&hash, generate_key_safe(dict, dict_head, LZSS_DICTSIZE), dict_head);

			dict_head = (dict_head + 1) % LZSS_DICTSIZE;
			waiting_bytes--;
		}
	}

	bitstream_finish(&bs);
	*outsize = bs.used_bytes;
	return bs.buffer;
}

static int
read_byte_mem(const unsigned char** ptrs)
{
	if (ptrs[0] >= ptrs[1])
		return -1;
	else {
		return *ptrs[0]++;
	}
}

unsigned char*
th_lz_mem(const unsigned char* in, unsigned int insize, unsigned int* outsize)
{
	const unsigned char* ptrs[] = { in, in + insize };
	return th_lz(outsize, (read_byte_fptr)read_byte_mem, &ptrs);
}

static int
read_byte_fd(FILE* fd)
{
	int c = fgetc(fd);

	if (c == EOF)
		return -1;
	else
		return c;
}

unsigned char*
th_lz_fd(FILE* in, unsigned int* outsize)
{
	return th_lz(outsize, (read_byte_fptr)read_byte_fd, in);
}
